from llama_index.storage.docstore.mongodb import MongoDocumentStore
from llama_index.core.storage.docstore.types import RefDocInfo
from typing import Optional, Dict
from dataclasses import asdict

class MongoDocumentStoreExtend(MongoDocumentStore):


    def get_all_doc(self, file_path: Optional[str] = None)  -> Optional[Dict[str, RefDocInfo]]:
        all_docs = self.get_all_ref_doc_info()
        if file_path is None:
            return all_docs

        target_docs = {}
        for doc_id, doc in all_docs.items():
            if doc.metadata.get("file_path", None) == file_path:
                target_docs[doc_id] = doc

        return target_docs
    
    
    def set_document_status(self, doc_id: str, status: str) -> None:
        ref_doc_info = self.get_ref_doc_info(doc_id)
        if ref_doc_info is None:
            raise ValueError(f"Document with id {doc_id} not found.")
        
        ref_doc_info.metadata["status"] = status
        
        self._kvstore.put(
            key=doc_id,
            val=asdict(ref_doc_info),
            collection=self._ref_doc_collection
        )

        pass


    def set_status(self, doc_ids: list[str], status: str) -> None:
        for doc_id in doc_ids:
            self.set_document_status( doc_id, status)



from pydantic import BaseModel, Field
from typing import Optional


class AWSS3Config(BaseModel):
    endpoint_url: str = Field(description="AWS S3 Endpoint URL")
    access_key: str = Field(description="AWS S3 Access Key")
    secret_key: str = Field(description="AWS S3 Secret Key")
    bucket: str = Field(description="AWS S3 Bucket Name")
    region: Optional[str] = Field(default="cn-north-1", description="AWS S3 Region")

class SQLDBConfig(BaseModel):
    prefix: str = Field(default="postgresql+asyncpg", description="Database URL prefix")
    user: str = Field(description="Database user")
    password: str = Field(description="Database password")
    database: str = Field(description="Database name")
    host: str = Field(description="Database host")
    port: int = Field(description="Database port")


class MongoDBConfig(BaseModel):
    user: str = Field(description="MongoDB user")
    password: str = Field(description="MongoDB password")
    host: str = Field(description="MongoDB host")
    port: int = Field(description="MongoDB port")
    database: str = Field(description="MongoDB database name")


class VectorDBConfig(BaseModel):
    host: str = Field(description="Vector DB host")
    port: int = Field(description="Vector DB port")
    collection_name: str = Field(description="Vector DB collection name")


class StorageConfig(BaseModel):
    file_store: Optional[AWSS3Config] = Field(description="File Store Configuration", default=None)
    doc_store: Optional[MongoDBConfig] = Field(description="Document Store Configuration", default=None)
    index_store: Optional[MongoDBConfig] = Field(description="Index Store Configuration", default=None)
    vector_store: Optional[VectorDBConfig] = Field(description="Vector Store Configuration", default=None)



from .storage import StorageConfig
from pydantic import BaseModel, Field
from typing import Optional

class LLMModelConfig(BaseModel):
    """
    for openai
    """
    model_name: str = Field(description="LLM model name")
    api_base: str = Field(description="API base URL for the LLM model")
    api_key: str = Field(description="API key for the LLM model")
    other_parameters: Optional[dict] = Field(default_factory=dict, description="Other parameters for the LLM model")


class RAGConfig(BaseModel):
    
    embedding_model: LLMModelConfig = Field(description="Embedding Model Configuration")
    generation_model: LLMModelConfig = Field(description="Generation Model Configuration")
    storage: StorageConfig = Field(description="Storage Configuration", default_factory=StorageConfig)


from .rag import RAGConfig
from pydantic import BaseModel, Field
from typing import List, Dict
import yaml


class ServerConfig(BaseModel):
    host: str = Field(description="Server host")
    port: int = Field(description="Server port")


class UserInfo(BaseModel):
    username: str = Field(description="Username")
    role: str = Field(description="User role")


class Config(BaseModel):
    server: ServerConfig = Field(description="Server Configuration")
    rag: RAGConfig = Field(description="RAG Configuration")
    user: Dict[str, UserInfo] = Field(description="User Information")

    @classmethod
    def from_yaml(cls, file_path: str) -> "Config":
        with open(file_path, 'r') as f:
            data = yaml.safe_load(f)
        return cls(**data)
    



from ..card.config import RAGConfig, LLMModelConfig

from llama_index.llms.openai_like import OpenAILike
from llama_index.embeddings.openai_like import OpenAILikeEmbedding
from llama_index.core import Settings
from llama_index.core.node_parser import SentenceSplitter

class RAGSetting:
    def __init__(self, config: RAGConfig):
        self.config = config


    def set_default_models(self):
        if self.config.embedding_model is not None:
            config: LLMModelConfig = self.config.embedding_model
            Settings.embed_model = OpenAILikeEmbedding(
                model_name=config.model_name,
                api_base=config.api_base,
                api_key=config.api_key,
                **config.other_parameters
            )

        if self.config.generation_model is not None:
            # MOST LIKELY NEED TO SETUP: is_chat_model=True
            config: LLMModelConfig = self.config.generation_model
            Settings.llm = OpenAILike(
                model_name=config.model_name,
                api_base=config.api_base,
                api_key=config.api_key,
                **config.other_parameters
            )


    def set_default_transformations(self):
        Settings.transformations = [
            SentenceSplitter(),
            Settings.embed_model
        ]

    def adjust_default_settings(self):
        self.set_default_models()
        self.set_default_transformations()


from ..card import Config, RetrievalData, UserInfo
from ..service import RAGService
from .deps import authenticate, authorize

import uvicorn
from fastapi import FastAPI, Depends
from contextlib import asynccontextmanager


class DEMORAG:
    def __init__(self, config_path: str):
        self.config = Config.from_yaml(config_path)

        self.app = self.__get_app()
        self.rag_service = None

    @asynccontextmanager
    async def lifespan(self, app: FastAPI):

        # Startup actions
        self.rag_service = RAGService(self.config.rag)
        self.rag_service.build()

        app.state.user_db = self.config.user

        yield

        # Shutdown actions
        print("Shutting down DEMORAG application.")

    
    def __get_app(self) -> FastAPI:

        app = FastAPI(title="Demo RAG Service API", lifespan=self.lifespan)


        @app.post("/retrieve")
        def retrieval(
            data: RetrievalData,
            user_info: UserInfo = Depends(authenticate), 
        ):
            """
            Retrieve relevant information based on the input text.
            """

            # step 1: get retrieval result
            result = self.rag_service.retrieve(data.text)

            # step 2: create response
            simple_result = self.rag_service.get_simple_result(result)

            return {"results": simple_result}
        
        @app.delete("/admin/clear")
        def clear(
            authorize: bool = Depends(authorize)
        ):
            """
            Clear all data in vector store, document store and index store.
            """
            self.rag_service.clear()
            return {"detail": "All data cleared successfully."}

        return app


    def run(self):
        uvicorn.run(
            app=self.app,
            host=self.config.server.host,
            port=self.config.server.port
        )


from fastapi import Request, HTTPException, Depends


def authenticate(request: Request) -> bool:
    """
    Verify the provided token against the expected token.
    """
    token = request.headers.get("Token")
    user_db = request.app.state.user_db

    if token not in user_db.keys():
        raise HTTPException(status_code=401, detail="Unauthorized: Invalid or missing token") 

    return user_db[token]

def authorize(user_info = Depends(authenticate)):
    """
    Authorize the user based on their role.
    """
    if user_info.role != "admin":
        raise HTTPException(status_code=403, detail="Forbidden: You do not have access to this resource")
    return True




from llama_index.storage.index_store.mongodb import MongoIndexStore
from llama_index.vector_stores.qdrant import QdrantVectorStore
from llama_index.core import StorageContext, SimpleDirectoryReader, VectorStoreIndex
from llama_index.core.schema import NodeWithScore
from qdrant_client import QdrantClient
from s3fs import S3FileSystem
from typing import List

from .rag_setting import RAGSetting
from .document_store_extend import MongoDocumentStoreExtend
from ..card.config import RAGConfig


class RAGService:
    def __init__(self, config: RAGConfig):
        self.config = config
        RAGSetting(self.config).adjust_default_settings()

        self.sync_fs = None
        self.storage_context = None
        self.index = None
        self.retriever = None
        self.query_engine = None

        self.__initialize_filestore()
        self.__initialize_storage_context()


    def __initialize_filestore(self):

        fs_config = self.config.storage.file_store
        self.sync_fs = S3FileSystem(
            key=fs_config.access_key,
            secret=fs_config.secret_key,
            client_kwargs={
                "endpoint_url": fs_config.endpoint_url,
                "region_name": fs_config.region,
                "verify": False,
                "use_ssl": False
            },
            config_kwargs={"s3": {"addressing_style": "path"}}
        )


    def __initialize_storage_context(self):

        doc_config = self.config.storage.doc_store
        doc_mongodb_uri = f"mongodb://{doc_config.user}:{doc_config.password}@{doc_config.host}:{doc_config.port}/{doc_config.database}?authSource=admin"
        document_store = MongoDocumentStoreExtend.from_uri(
            uri=doc_mongodb_uri,
            db_name=doc_config.database,
            namespace="docstore",
        )

        index_config = self.config.storage.index_store
        index_mongodb_uri = f"mongodb://{index_config.user}:{index_config.password}@{index_config.host}:{index_config.port}/{index_config.database}?authSource=admin"

        index_store = MongoIndexStore.from_uri(
            uri=index_mongodb_uri,
            db_name=index_config.database,
            namespace="indexstore"
        )

        vector_config = self.config.storage.vector_store
        vector_store = QdrantVectorStore(
            client=QdrantClient(
                host=vector_config.host, 
                port=vector_config.port
            ), 
            collection_name=vector_config.collection_name
        )

        # NOTE: VERY IMPORTANT FOR THE INDEX STUCTURE CONSISTENCY
        vector_store.stores_text = False

        self.storage_context = StorageContext.from_defaults(
            docstore=document_store,
            index_store=index_store,
            vector_store=vector_store
        )


    def __set_file_metadata(self, file_path: str) -> dict:
        metadata = {
            "file_path": file_path,
            "status": "processing",
        }
        return metadata


    def build(self):

        # step 1: load data from filestore
        reader = SimpleDirectoryReader(
            input_dir=self.config.storage.file_store.bucket,
            fs=self.sync_fs,
            recursive=True,
            file_metadata=self.__set_file_metadata
        )

        documents = reader.load_data()

        # step 2: build a new index
        self.index = VectorStoreIndex.from_documents(
            documents,
            storage_context=self.storage_context
        )
        self.retriever = self.index.as_retriever(similarity_top_k=1)
        self.query_engine = self.index.as_query_engine()

        # step 3: set status to active
        self.storage_context.docstore.set_status(
            [doc.doc_id for doc in documents],
            "active"
        )


    def clear(self, index=True):

        """
        clear all data in vector store, document store and index store
        """
        self.storage_context.vector_store.clear()
        
        ref_doc_infos = self.storage_context.docstore.get_all_ref_doc_info()
        for doc_id in ref_doc_infos.keys():
            self.storage_context.docstore.delete_ref_doc(doc_id)

        if index:
            index_structs = self.storage_context.index_store.index_structs()
            for struct in index_structs:
                self.storage_context.index_store.delete_index_struct(struct.index_id)


    def retrieve(self, text: str):

        # step 1: retrieve from index
        result = self.retriever.retrieve(text)

        # step 2: filter the documents with status "active"
        filtered_results = []
        for node_with_score in result:
            ref_doc_id = node_with_score.node.ref_doc_id
            ref_doc_info = self.storage_context.docstore.get_ref_doc_info(ref_doc_id)
            if ref_doc_info and ref_doc_info.metadata.get("status", "") == "active":
                filtered_results.append(node_with_score)

        return filtered_results
    
    def get_simple_result(self, result: List[NodeWithScore]):
        simple_results = []
        for node_with_score in result:
            simple_results.append({
                "content": node_with_score.node.get_content(),
                "score": node_with_score.score
            })
        return simple_results

    def upload_file(self, data, file_path: str):
        with self.sync_fs.open(file_path, 'wb') as f:
            f.write(data)


    def download_file(self, file_path: str):
        with self.sync_fs.open(file_path, 'rb') as f:
            data = f.read()
        return data
    

    def delete_file(self, file_path: str):
        self.sync_fs.rm(file_path)


    def add(self, file_path: str):

        # step 1: load the file
        reader = SimpleDirectoryReader(
            input_files=[file_path],
            fs=self.sync_fs,
            recursive=True,
            file_metadata=self.__set_file_metadata
        )

        documents = reader.load_data()

        # step 2: insert to index and storecontext
        for doc in documents:
            self.index.insert(document=doc)

        # step 3: set status to active
        self.storage_context.docstore.set_status(
            [doc.doc_id for doc in documents],
            "active"
        )

    def delete(self, file_path: str):

        # step 1: get the target docs
        target_docs = self.storage_context.docstore.get_all_doc(file_path=file_path)

        # step 2: set status to deleted
        self.storage_context.docstore.set_status(
            [doc_id for doc_id in target_docs.keys()],
            "deleted"
        )

        # step 3: delete from index
        for doc_id in target_docs.keys():
            self.index.delete_ref_doc(ref_doc_id=doc_id, delete_from_docstore=True)
